import hashlib
import tempfile
import os
from typing import Annotated

from gridfs import NoFile
from fastapi import APIRouter, Depends, File, Form, HTTPException, UploadFile, status
from fastapi.responses import FileResponse

from pytupli.schema import (
    ArtifactMetadata,
    ArtifactMetadataItem,
    UserOut,
    BaseFilter,
    RESOURCE_TYPE,
    RIGHT,
)
from pytupli.server.api.dependencies import get_db_handler
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.config import (  # environment variables, constants and Handler Factory
    ARTIFACTS_COLLECTION_NAME,
)
from pytupli.server.management.authorization import (
    check_authorization,
    check_group_permissions,
    inject_read_permission_filter,
)

router = APIRouter()

filename_template = 'artifact_{id}'


async def _download_artifact_data(
    db_handler: MongoDBHandler, artifact_id: str
) -> tuple[bytes, dict]:
    """Download artifact data and metadata from database."""
    try:
        data, metadata = await db_handler.download_file(
            ARTIFACTS_COLLECTION_NAME, {'metadata.id': artifact_id}
        )
        return data, metadata
    except NoFile as e:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f'Could not find file {artifact_id}: {e}'
        )


async def _update_artifact_published_in(
    db_handler: MongoDBHandler, artifact_id: str, published_in_list: list[str]
) -> None:
    """Update artifact's published_in list by re-uploading with new metadata."""
    try:
        # Download current file and metadata
        data, metadata = await _download_artifact_data(db_handler, artifact_id)

        # Update the published_in field in metadata
        metadata['published_in'] = published_in_list

        # Delete old file
        await db_handler.delete_file(ARTIFACTS_COLLECTION_NAME, {'metadata.id': artifact_id})

        # Re-upload with updated metadata
        await db_handler.upload_file(
            ARTIFACTS_COLLECTION_NAME,
            data,
            filename_template.format(id=metadata['id']),
            metadata=metadata,
        )
    except Exception as e:
        if 'Multiple Objects with id' in str(e):
            raise HTTPException(status_code=status.HTTP_300_MULTIPLE_CHOICES, detail=str(e))
        else:
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e))


@router.post('/upload')
async def artifact_upload(
    data: Annotated[UploadFile, File()],
    metadata: Annotated[
        str, Form()
    ],  # this must conform to DataSourceMetadata but can only be a str in a multipart form
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.ARTIFACT, RIGHT.ARTIFACT_CREATE)
    ),
) -> ArtifactMetadataItem:
    user, _ = auth_result
    metadata: ArtifactMetadata = ArtifactMetadata.model_validate_json(metadata)

    contents = await data.read()
    try:
        # Hash the raw file contents combined with the username
        file_hash = hashlib.sha256(contents + user.username.encode('utf-8')).hexdigest()

        artifact_metadata = ArtifactMetadataItem.create_new(
            created_by=user.username,
            hash=file_hash,
            **metadata.model_dump(),
        )

        # Write the file to MongoDB using GridFS
        object_metadata = await db_handler.upload_file(
            ARTIFACTS_COLLECTION_NAME,
            contents,
            filename_template.format(id=artifact_metadata.id),
            artifact_metadata.model_dump(),
        )
        return object_metadata
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to save file: {e}',
        )


@router.put('/publish')
async def artifact_publish(
    artifact_id: str,
    publish_in: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, ArtifactMetadataItem | None] = Depends(
        check_authorization(RESOURCE_TYPE.ARTIFACT, RIGHT.ARTIFACT_READ, requires_ownership=True)
    ),
) -> None:
    user, artifact = auth_result

    if not artifact:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Artifact does not exist')

    # Check if user has ARTIFACT_CREATE rights in the specified group
    await check_group_permissions(db_handler, user, publish_in, [RIGHT.ARTIFACT_CREATE])

    # Check if already published in this group
    current_published_in = artifact.published_in
    if publish_in in current_published_in:
        return

    # Add the new group to published_in list
    updated_published_in = current_published_in + [publish_in]
    await _update_artifact_published_in(db_handler, artifact_id, updated_published_in)


@router.put('/unpublish')
async def artifact_unpublish(
    artifact_id: str,
    unpublish_from: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, ArtifactMetadataItem | None] = Depends(
        check_authorization(RESOURCE_TYPE.ARTIFACT, RIGHT.ARTIFACT_READ)
    ),
) -> None:
    user, artifact = auth_result

    if not artifact:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Artifact does not exist')

    # Check if user has CONTENT_DELETE rights in the specified group
    await check_group_permissions(db_handler, user, unpublish_from, [RIGHT.ARTIFACT_DELETE])

    # Check if currently published in this group
    current_published_in = artifact.published_in
    if unpublish_from not in current_published_in:
        return

    # Remove the group from published_in list
    updated_published_in = [group for group in current_published_in if group != unpublish_from]
    await _update_artifact_published_in(db_handler, artifact_id, updated_published_in)


@router.post('/list')
async def artifact_list(
    filter: BaseFilter = BaseFilter(),
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(check_authorization(RESOURCE_TYPE.ARTIFACT)),
) -> list[ArtifactMetadataItem]:
    user, _ = auth_result
    filter.apply_prefix('metadata')
    filter = await inject_read_permission_filter(
        filter, user, RESOURCE_TYPE.ARTIFACT, db_handler, prefix_path='metadata'
    )
    query_filter = db_handler.convert_filter_to_query(filter) if filter else {}
    try:
        files = await db_handler.download_files(ARTIFACTS_COLLECTION_NAME, query_filter)
        metadata = []
        for file in files:
            metadata.append(ArtifactMetadataItem(**file['metadata']))
        return metadata
    except Exception as e:
        raise HTTPException(status_code=500, detail=f'Failed to list arifact: {str(e)}')


@router.get('/download')
async def artifact_download(
    artifact_id: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, ArtifactMetadataItem | None] = Depends(
        check_authorization(RESOURCE_TYPE.ARTIFACT, RIGHT.ARTIFACT_READ)
    ),
) -> FileResponse:
    _, artifact = auth_result
    if not artifact:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Artifact does not exist')

    try:
        # Get both data and metadata (resource already verified by authorization)
        data, metadata = await db_handler.download_file(
            ARTIFACTS_COLLECTION_NAME, {'metadata.id': artifact_id}
        )

        metadata = ArtifactMetadataItem(**metadata)

        # Create a temporary file to serve with FileResponse
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        try:
            temp_file.write(data)
            temp_file.flush()
            temp_file_path = temp_file.name

            # Prepare headers with metadata
            headers = {'X-Metadata': metadata.model_dump_json()}

            return FileResponse(
                path=temp_file_path,
                headers=headers,
                filename=metadata.name,
                background=lambda: os.unlink(
                    temp_file_path
                ),  # Clean up the temp file after sending
            )

        except Exception as e:
            # Clean up temp file in case of errors
            os.unlink(temp_file.name)
            raise e
    except FileNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except Exception as e:
        if 'Multiple Objects with id' in str(e):
            raise HTTPException(status_code=status.HTTP_300_MULTIPLE_CHOICES, detail=str(e))
        else:
            raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e))


@router.delete('/delete')
async def artifact_delete(
    artifact_id: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    _: tuple[UserOut, dict] = Depends(
        check_authorization(RESOURCE_TYPE.ARTIFACT, RIGHT.ARTIFACT_DELETE)
    ),
) -> None:
    try:
        # Delete the artifact (existence already verified by authorization)
        _ = await db_handler.delete_file(ARTIFACTS_COLLECTION_NAME, {'metadata.id': artifact_id})
        # if database returns delete_count 0, then the file was not found
        # if r.deleted_count == 0:
        #    raise FileNotFoundError(f'Artifact with id {artifact_id} not found')
        return
    except FileNotFoundError as e:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(e))
    except NoFile as e:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f'Could not find file {artifact_id}: {e}'
        )
    except Exception as e:
        if 'Multiple Objects with id' in str(e):
            raise HTTPException(status_code=status.HTTP_300_MULTIPLE_CHOICES, detail=str(e))
        else:
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f'Could not delete artifact {artifact_id}: {e}',
            )



import logging

from fastapi import APIRouter, Depends, HTTPException, status

from pytupli.server.api.dependencies import get_db_handler
from pytupli.schema import UserRole, UserOut, RESOURCE_TYPE, RIGHT
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.config import USER_COLLECTION_NAME, USER_ROLES_COLLECTION_NAME
from pytupli.server.management.authorization import check_authorization

router = APIRouter()
logging.getLogger('passlib').setLevel(logging.ERROR)


@router.post('/create')
async def create_role(
    role: UserRole,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    _: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.ROLE, RIGHT.ROLE_MANAGEMENT)
    ),
) -> UserRole:
    """Create a new role."""
    try:
        # check if role already exists
        query = {'role': role.role}
        role_entry = await db_handler.get_item(USER_ROLES_COLLECTION_NAME, query)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to create role: {str(e)}',
        )

    if role_entry:
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail='Role already exists')

    try:
        # create the role in the db
        await db_handler.create_item(USER_ROLES_COLLECTION_NAME, role.model_dump())
        return role
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to create role: {str(e)}',
        )


@router.get('/list')
async def list_roles(
    db_handler: MongoDBHandler = Depends(get_db_handler),
    _: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.ROLE, RIGHT.ROLE_MANAGEMENT)
    ),
) -> list[UserRole]:
    """List all roles."""
    try:
        roles = await db_handler.get_items(USER_ROLES_COLLECTION_NAME, {})
        role_objects = [UserRole(**role) for role in roles]
        return role_objects
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to list roles: {str(e)}',
        )


@router.delete('/delete')
async def delete_role(
    role_name: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    _: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.ROLE, RIGHT.ROLE_MANAGEMENT)
    ),
) -> None:
    """Delete a role."""
    try:
        # check if role exists
        role_query = {'role': role_name}
        role_entry = await db_handler.get_item(USER_ROLES_COLLECTION_NAME, role_query)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to delete role: {str(e)}',
        )

    if not role_entry:
        return

    try:
        # Check if any users have this role and remove it
        users = await db_handler.get_items(USER_COLLECTION_NAME, {})
        for user in users:
            user_obj = UserOut(**user)
            updated_memberships = []
            modified = False

            for membership in user_obj.memberships:
                updated_roles = [role for role in membership.roles if role != role_name]
                if len(updated_roles) != len(membership.roles):
                    modified = True
                    # Only keep membership if it still has roles
                    if updated_roles:
                        membership.roles = updated_roles
                        updated_memberships.append(membership)
                else:
                    updated_memberships.append(membership)

            if modified:
                # Update user if memberships changed
                update = {'$set': {'memberships': [m.model_dump() for m in updated_memberships]}}
                await db_handler.update_item(
                    USER_COLLECTION_NAME, {'username': user_obj.username}, update
                )

        # Delete the role
        await db_handler.delete_item(USER_ROLES_COLLECTION_NAME, role_query)
        return

    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to delete role: {str(e)}',
        )



import logging

from fastapi import APIRouter, Depends, HTTPException, status

from pytupli.server.api.dependencies import get_db_handler
from pytupli.schema import (
    DEFAULT_ROLE,
    BaseFilter,
    FilterType,
    Group,
    GroupMembershipQuery,
    GroupWithMembers,
    Membership,
    UserOut,
    RESOURCE_TYPE,
    RIGHT,
)
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.config import (
    GROUPS_COLLECTION_NAME,
    USER_COLLECTION_NAME,
    USER_ROLES_COLLECTION_NAME,
)
from pytupli.server.management.authorization import (
    check_authorization,
    get_user_memberships,
    get_rights_in_groups,
)

router = APIRouter()
logging.getLogger('passlib').setLevel(logging.ERROR)


@router.post('/create')
async def create_group(
    group: Group,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.GROUP, RIGHT.GROUP_CREATE)
    ),
) -> Group:
    """Create a new group."""
    current_user, _ = auth_result
    try:
        # check if group name is already taken
        query = {'name': group.name}
        group_entry = await db_handler.get_item(GROUPS_COLLECTION_NAME, query)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to create group: {str(e)}',
        )

    if group_entry:
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail='Group already exists')

    try:
        # create the group in the db
        await db_handler.create_item(GROUPS_COLLECTION_NAME, group.model_dump())

        # add admin membership for the creator
        creator_membership = Membership(
            group=group.name, roles=[DEFAULT_ROLE.GROUP_ADMIN, DEFAULT_ROLE.CONTRIBUTOR]
        )

        # Update user memberships
        updated_memberships = current_user.memberships + [creator_membership]
        update = {'$set': {'memberships': [m.model_dump() for m in updated_memberships]}}
        await db_handler.update_item(
            USER_COLLECTION_NAME, {'username': current_user.username}, update
        )

        return group
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to create group: {str(e)}',
        )


@router.get('/list')
async def list_groups(
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(check_authorization(RESOURCE_TYPE.GROUP)),
) -> list[Group]:
    """List all groups based on user's access rights."""
    current_user, _ = auth_result

    try:
        # Check if user has global GROUP_READ right
        global_rights = await get_rights_in_groups(db_handler, current_user, ['global'])
        has_global_group_read = RIGHT.GROUP_READ in global_rights

        if has_global_group_read:
            # Get all groups
            accessible_groups = await db_handler.get_items(GROUPS_COLLECTION_NAME, {})
        else:
            # User can only see groups they are a member of
            user_memberships = get_user_memberships(current_user)
            filter = BaseFilter(
                type=FilterType.OR,
                filters=[
                    BaseFilter(field='name', value=membership.group)
                    for membership in user_memberships
                ],
            )
            accessible_groups = await db_handler.query_items(GROUPS_COLLECTION_NAME, filter)

        return [Group(**group) for group in accessible_groups]

    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to list groups: {str(e)}',
        )


@router.get('/read')
async def read_group(
    group_name: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.GROUP, RIGHT.GROUP_READ)
    ),
) -> GroupWithMembers:
    """Read a specific group with its members if user has appropriate access."""
    current_user, _ = auth_result

    try:
        group = await db_handler.get_item(GROUPS_COLLECTION_NAME, {'name': group_name})
        if not group:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Group not found')

        group = Group(**group)

        # Check if user has global USER_READ right
        global_rights = await get_rights_in_groups(db_handler, current_user, ['global', group.name])
        has_group_read = RIGHT.USER_READ in global_rights

        # Initialize the group response
        group_with_members = GroupWithMembers(**group.model_dump(), members=[])

        # If user has USER_READ in the group, include members
        # This could be made more efficient by expanding the query concept
        if has_group_read:
            # Get all users and filter those who are members of this group
            all_users = await db_handler.get_items(USER_COLLECTION_NAME, {})

            for user_data in all_users:
                user_obj = UserOut(**user_data)
                # Check if user is a member of this group
                for membership in user_obj.memberships:
                    if membership.group == group_name:
                        group_with_members.members.append(user_obj.username)
                        break

        return group_with_members

    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to read group: {str(e)}',
        )


@router.delete('/delete')
async def delete_group(
    group_name: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    _: tuple[UserOut, None] = Depends(check_authorization(RESOURCE_TYPE.GROUP, RIGHT.GROUP_DELETE)),
) -> None:
    """Delete a group."""

    group = await db_handler.get_item(GROUPS_COLLECTION_NAME, {'name': group_name})
    if not group:
        return

    try:
        # Remove group membership from all users
        # This could be made more efficient by expanding the query concept
        users = await db_handler.get_items(USER_COLLECTION_NAME, {})
        for user in users:
            user_obj = UserOut(**user)
            updated_memberships = [
                membership for membership in user_obj.memberships if membership.group != group_name
            ]
            if len(updated_memberships) != len(user_obj.memberships):
                # Update user if memberships changed
                update = {'$set': {'memberships': [m.model_dump() for m in updated_memberships]}}
                await db_handler.update_item(
                    USER_COLLECTION_NAME, {'username': user_obj.username}, update
                )

        # Delete the group
        await db_handler.delete_item(GROUPS_COLLECTION_NAME, {'name': group_name})
        return

    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to delete group: {str(e)}',
        )


@router.post('/add-members')
async def add_members(
    membership_query: GroupMembershipQuery,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.GROUP, RIGHT.GROUP_UPDATE)
    ),
) -> GroupWithMembers:
    """Add members to a group with specified roles."""
    current_user, _ = auth_result

    group = await db_handler.get_item(GROUPS_COLLECTION_NAME, {'name': membership_query.group_name})
    if not group:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Group not found')

    group = Group(**group)

    # Check if user has global USER_READ right (otherwise they can't add users)
    global_user_rights = await get_rights_in_groups(db_handler, current_user, ['global'])
    has_user_read = RIGHT.USER_READ in global_user_rights

    if not has_user_read:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail='You do not have sufficient rights to add users to this group',
        )

    # Get current user's rights to validate they can assign these roles
    current_user_rights = await get_rights_in_groups(
        db_handler, current_user, ['global', membership_query.group_name]
    )

    # Validate that current user has rights to assign all requested roles
    try:
        for member in membership_query.members:
            for role in member.roles:
                role_entry = await db_handler.get_item(USER_ROLES_COLLECTION_NAME, {'role': role})

                if not role_entry:
                    raise HTTPException(
                        status_code=status.HTTP_404_NOT_FOUND, detail=f'Role {role} not found'
                    )

                role_rights = set(role_entry.get('rights', []))
                if not role_rights.issubset(current_user_rights):
                    raise HTTPException(
                        status_code=status.HTTP_403_FORBIDDEN,
                        detail=f'You do not have sufficient rights to assign role {role}',
                    )

    except HTTPException as e:
        raise e
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to validate role assignments: {str(e)}',
        )

    # Add members to the group
    try:
        for member in membership_query.members:
            if not member.roles:
                continue  # Skip if no roles are specified

            # Check if user exists
            user_entry = await db_handler.get_item(USER_COLLECTION_NAME, {'username': member.user})
            if not user_entry:
                raise HTTPException(
                    status_code=status.HTTP_404_NOT_FOUND, detail=f'User {member.user} not found'
                )

            user_obj = UserOut(**user_entry)

            # Remove existing membership in this group (if any)
            updated_memberships = [
                membership for membership in user_obj.memberships if membership.group != group.name
            ]

            # Add new membership
            new_membership = Membership(group=group.name, roles=member.roles)
            updated_memberships.append(new_membership)

            # Update user
            update = {'$set': {'memberships': [m.model_dump() for m in updated_memberships]}}
            await db_handler.update_item(USER_COLLECTION_NAME, {'username': member.user}, update)

        return await read_group(
            group_name=membership_query.group_name,
            db_handler=db_handler,
            auth_result=auth_result,
        )

    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to add members: {str(e)}',
        )


@router.post('/remove-members')
async def remove_members(
    membership_query: GroupMembershipQuery,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.GROUP, RIGHT.GROUP_UPDATE)
    ),
) -> GroupWithMembers:
    """Remove members from a group."""
    current_user, _ = auth_result

    group = await db_handler.get_item(GROUPS_COLLECTION_NAME, {'name': membership_query.group_name})
    if not group:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Group not found')

    group = Group(**group)

    # Check if user has USER_READ rights
    current_user_rights = await get_rights_in_groups(
        db_handler, current_user, ['global', group.name]
    )
    has_user_read = RIGHT.USER_READ in current_user_rights

    if not has_user_read:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail='You do not have sufficient rights to remove users from this group',
        )

    # Remove members from the group
    try:
        for member in membership_query.members:
            # Check if user exists
            user_entry = await db_handler.get_item(USER_COLLECTION_NAME, {'username': member.user})
            if not user_entry:
                continue  # Skip if user doesn't exist

            user_obj = UserOut(**user_entry)

            # Remove membership from this group
            updated_memberships = [
                membership for membership in user_obj.memberships if membership.group != group.name
            ]

            # Only update if membership was actually removed
            if len(updated_memberships) < len(user_obj.memberships):
                update = {'$set': {'memberships': [m.model_dump() for m in updated_memberships]}}
                await db_handler.update_item(
                    USER_COLLECTION_NAME, {'username': member.user}, update
                )

        return await read_group(
            group_name=membership_query.group_name,
            db_handler=db_handler,
            auth_result=auth_result,
        )

    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to remove members: {str(e)}',
        )



# JWT logic following the official FastAPI documentation https://fastapi.tiangolo.com/tutorial/security/oauth2-jwt/
import os
from datetime import datetime, timedelta, timezone

import jwt
from fastapi import Depends, HTTPException, Request, Security, status
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from jwt.exceptions import InvalidTokenError
from passlib.context import CryptContext

from pytupli.server.api.dependencies import get_db_handler
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.config import (  # environment variables, constants and Handler Factory
    BENCHMARK_COLLECTION_NAME,
    ARTIFACTS_COLLECTION_NAME,
    EPISODES_COLLECTION_NAME,
    USER_COLLECTION_NAME,
    OPEN_ACCESS_MODE,
    OPEN_SIGNUP_MODE,
    USER_ROLES_COLLECTION_NAME,
)
from pytupli.schema import (
    RESOURCE_TYPE,
    ArtifactMetadataItem,
    BaseFilter,
    BenchmarkHeader,
    DBItem,
    EpisodeHeader,
    FilterType,
    Membership,
    UserOut,
    UserRole,
    RIGHT,
    RIGHT_BUNDLE_CONTENT_CREATE,
    RIGHT_BUNDLE_CONTENT_DELETE,
    RIGHT_BUNDLE_CONTENT_READ,
    DEFAULT_ROLE,
)

http_bearer = HTTPBearer(auto_error=False)

SECRET_KEY = None  # Initialize as None to avoid errors at import time
ALGORITHM = 'HS256'

pwd_context = CryptContext(schemes=['bcrypt'], deprecated='auto')


AUTH_ERROR = HTTPException(
    status_code=status.HTTP_403_FORBIDDEN,
    detail='You do not have sufficient permissions for this operation.',
)

GUEST_USER = UserOut(
    username='$$GUEST$$', memberships=[Membership(group='global', roles=[DEFAULT_ROLE.GUEST])]
)


def initialize_secret_key():
    """
    Initializes secret key needed for token creation and validation.
    """
    global SECRET_KEY
    SECRET_KEY = os.getenv('API_SECRET_KEY')
    if SECRET_KEY is None:
        raise ValueError('API_SECRET_KEY environment variable must be set!')


def hash_password(password: str) -> str:
    """
    Hashes a password using bcrypt.
    """
    return pwd_context.hash(password)  # is automatically salted


def verify_password(plain_password: str, hashed_password: str) -> bool:
    """
    Verifies a password against a hashed password.\n
    Returns True if the password matches, False otherwise.
    """
    return pwd_context.verify(plain_password, hashed_password)


def create_token(data: dict, expires_delta: timedelta = timedelta(minutes=15)) -> str:
    """
    Function to create an access or refresh token.
    Args:
        data (dict): The data to be encoded in the token.
        expires_delta (timedelta): The expiration time of the token.
            Default is 15 minutes.
    Returns:
        str: The generated token.
    """
    to_encode = data.copy()
    # set exiration time of token
    expire = datetime.now(timezone.utc) + expires_delta

    to_encode.update({'exp': expire})
    # create token
    encoded_jwt = jwt.encode(to_encode, SECRET_KEY, algorithm=ALGORITHM)
    return encoded_jwt


async def authenticate_user(
    db_handler: MongoDBHandler,
    credentials: HTTPAuthorizationCredentials = Security(http_bearer),
) -> UserOut | None:
    # Do we allow open access?
    if credentials is None and not OPEN_ACCESS_MODE:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail='You need to be authenticated with Bearer scheme to access the platform.',
            headers={'WWW-Authenticate': 'Bearer'},
        )

    if credentials is None and OPEN_ACCESS_MODE:
        return None

    if credentials is not None:
        # check that the authentication scheme is Bearer
        if not credentials.scheme == 'Bearer':
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='Invalid authentication scheme',
                headers={'WWW-Authenticate': 'Bearer'},
            )

    token = credentials.credentials

    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail='Could not validate credentials',
        headers={'WWW-Authenticate': 'Bearer'},
    )

    try:
        # decode token
        payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
        token_username: str = payload.get('issuer')
        # if "issuer" is not in payload, token is invalid as it does not come from our token creation logic
        if token_username is None:
            raise credentials_exception
    except InvalidTokenError:
        raise credentials_exception
    # get user from database
    user_dict = await db_handler.get_item(USER_COLLECTION_NAME, {'username': token_username})
    if user_dict is None:
        # No user found -> guest user
        return None

    return UserOut(**user_dict)


async def get_resource(
    resource_type: RESOURCE_TYPE, request_params: dict[str, str], db_handler: MongoDBHandler
) -> DBItem | None:
    """
    Returns a filter to get the resource by its type and id.
    """

    resource_id: str | None = request_params.get(resource_type.value.lower() + '_id')

    if resource_type == RESOURCE_TYPE.ARTIFACT:
        _, entry = await db_handler.download_file(
            ARTIFACTS_COLLECTION_NAME, {'metadata.id': resource_id}
        )
        entry = ArtifactMetadataItem(**entry) if entry else None
    elif resource_type == RESOURCE_TYPE.BENCHMARK:
        entry = await db_handler.get_item(BENCHMARK_COLLECTION_NAME, {'id': resource_id})
        entry = BenchmarkHeader(**entry) if entry else None
    elif resource_type == RESOURCE_TYPE.EPISODE:
        entry = await db_handler.get_item(EPISODES_COLLECTION_NAME, {'id': resource_id})
        entry = EpisodeHeader(**entry) if entry else None
    return entry


def get_user_memberships(user: UserOut) -> list[Membership]:
    """
    Returns the memberships of the user.
    """
    memberships = user.memberships.copy()

    if user.username == '$$GUEST$$':
        return memberships

    # Add default memberships for global and user-specific groups
    memberships.append(
        Membership(group='global', roles=[DEFAULT_ROLE.GLOBAL_MEMBER, DEFAULT_ROLE.CONTRIBUTOR])
    )
    memberships.append(
        Membership(group=user.username, roles=[DEFAULT_ROLE.USER_ADMIN, DEFAULT_ROLE.CONTRIBUTOR])
    )

    return memberships


async def get_rights_in_groups(
    db_handler: MongoDBHandler,
    user: UserOut,
    groups: list[str],
):
    user_memberships: list[Membership] = get_user_memberships(user)

    # Check if the published_in list contains any of the user's memberships and construct set of rights
    user_roles = set()
    user_rights = set()
    for membership in user_memberships:
        if membership.group in groups:
            user_roles.update(membership.roles)

    for role in user_roles:
        role_entry = await db_handler.get_item(USER_ROLES_COLLECTION_NAME, {'role': role})
        if role_entry:
            user_rights.update(role_entry.get('rights', []))

    return user_rights


async def _extract_request_params(request: Request) -> dict[str, str]:
    """Extract parameters from both query params and request body."""
    params = {}

    # Get all query parameters
    for key, value in request.query_params.items():
        params[key] = value

    # Try to get parameters from request body
    try:
        import json

        body = await request.body()
        if body:
            body_data = json.loads(body)
            if isinstance(body_data, dict):
                params.update(body_data)
    except Exception:
        # If we can't parse the body, continue with just query params
        pass

    return params


def check_authorization(
    resource_type: RESOURCE_TYPE,
    required_right: RIGHT | None = None,
    requires_ownership: bool = False,
):
    async def _check_authorization(
        request: Request,
        db_handler: MongoDBHandler = Depends(get_db_handler),
        credentials: HTTPAuthorizationCredentials = Security(http_bearer),
    ) -> tuple[UserOut, DBItem | None]:
        # Extract parameters from request (query params + body)
        request_params = await _extract_request_params(request)

        # Special rule for signup
        if (
            OPEN_SIGNUP_MODE
            and resource_type == RESOURCE_TYPE.USER
            and required_right == RIGHT.USER_CREATE
        ):
            return GUEST_USER, None

        user: UserOut | None = await authenticate_user(db_handler, credentials)
        if not user:
            user = GUEST_USER

        if not required_right:  # no specific right required
            return user, None

        user_rights = set()
        resource = None

        if resource_type == RESOURCE_TYPE.ROLE:
            # Role management requires global admin rights
            user_rights = await get_rights_in_groups(db_handler, user, ['global'])

        if resource_type in [RESOURCE_TYPE.USER, RESOURCE_TYPE.GROUP]:
            # creating users/groups requires create rights on global scope
            if required_right in [RIGHT.USER_CREATE, RIGHT.GROUP_CREATE]:
                user_rights = await get_rights_in_groups(db_handler, user, ['global'])

            if required_right in [RIGHT.GROUP_DELETE, RIGHT.GROUP_READ, RIGHT.GROUP_UPDATE]:
                # for group management
                group_name = request_params.get('group_name')
                user_rights = await get_rights_in_groups(db_handler, user, ['global', group_name])

            if required_right in [RIGHT.USER_READ, RIGHT.USER_UPDATE, RIGHT.USER_DELETE]:
                # for user management
                if user.username == request_params.get('username'):
                    # user is accessing their own data
                    user_rights = await get_rights_in_groups(
                        db_handler, user, ['global', user.username]
                    )
                else:
                    # otherwise only global admin can access other users
                    user_rights = await get_rights_in_groups(db_handler, user, ['global'])

        if resource_type in [
            RESOURCE_TYPE.ARTIFACT,
            RESOURCE_TYPE.BENCHMARK,
            RESOURCE_TYPE.EPISODE,
        ]:
            if required_right in RIGHT_BUNDLE_CONTENT_CREATE:
                user_rights = await get_rights_in_groups(
                    db_handler, user, ['global', user.username]
                )

            elif required_right in RIGHT_BUNDLE_CONTENT_READ + RIGHT_BUNDLE_CONTENT_DELETE:
                resource: DBItem | None = await get_resource(
                    resource_type, request_params, db_handler
                )
                if resource is None:
                    return user, None

                has_ownership = resource.created_by == user.username

                if requires_ownership and not has_ownership:
                    raise AUTH_ERROR

                if has_ownership:
                    user_rights = set(RIGHT_BUNDLE_CONTENT_DELETE + RIGHT_BUNDLE_CONTENT_READ)
                else:
                    if required_right in RIGHT_BUNDLE_CONTENT_READ:
                        # read rights in published_in groups
                        user_rights = await get_rights_in_groups(
                            db_handler, user, [*resource.published_in]
                        )
                    elif required_right in RIGHT_BUNDLE_CONTENT_DELETE:
                        # only global admin or creator can delete
                        user_rights = await get_rights_in_groups(db_handler, user, ['global'])

        if required_right not in user_rights:
            raise AUTH_ERROR

        return user, resource

    return _check_authorization


async def check_group_permissions(
    db_handler: MongoDBHandler, user: UserOut, group_name: str, required_rights: list[RIGHT]
) -> None:
    """Check if user has required rights in the specified group."""
    user_rights = await get_rights_in_groups(db_handler, user, [group_name])

    # Check if user has any of the required rights
    has_permission = any(right in user_rights for right in required_rights)

    if not has_permission:
        rights_str = ' or '.join([right.value for right in required_rights])
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=f'Insufficient permissions in group: {group_name}. Required: {rights_str}',
        )


async def inject_read_permission_filter(
    filter: BaseFilter,
    user: UserOut,
    resource_type: RESOURCE_TYPE,
    db_handler: MongoDBHandler,
    prefix_path: str = '',
) -> BaseFilter | None:
    """
    Function to inject read permission filter into the user-provided filter.

    Args:
        filter (BaseFilter): The user-provided filter.
        user (User): The user object.
        db_handler (MongoDBHandler): The database handler (to fetch user rights).
        prefix_path (str, optional): Prefix path to the fields relevant for permissions, e.g. "metadata".
            Defaults to "".

    Returns:
        BaseFilter | None: The modified filter or None if no filter is needed.
    """

    # Determine the appropriate read right based on resource type
    match resource_type:
        case RESOURCE_TYPE.ARTIFACT:
            resource_read_right = RIGHT.ARTIFACT_READ
        case RESOURCE_TYPE.BENCHMARK:
            resource_read_right = RIGHT.BENCHMARK_READ
        case RESOURCE_TYPE.EPISODE:
            resource_read_right = RIGHT.EPISODE_READ
        case RESOURCE_TYPE.USER:
            resource_read_right = RIGHT.USER_READ
        case RESOURCE_TYPE.GROUP:
            resource_read_right = RIGHT.GROUP_READ
        case _:
            raise ValueError(f'Unsupported resource type: {resource_type}')

    # In which groups does the user have read rights?
    user_roles = {
        membership.group: role
        for membership in get_user_memberships(user)
        for role in membership.roles
    }
    user_read_groups = []
    for group, role in user_roles.items():
        role_entry: dict = await db_handler.get_item(USER_ROLES_COLLECTION_NAME, {'role': role})
        if role_entry:
            role_obj = UserRole(**role_entry)
            if resource_read_right in role_obj.rights:
                user_read_groups.append(group)

    prefix_path = prefix_path + '.' if prefix_path else ''
    allowed_filter = BaseFilter(type='IN', key=f'{prefix_path}published_in', value=user_read_groups)

    # Inject allowed filter into the user-provided filter
    if filter:
        if allowed_filter:
            filter = BaseFilter(
                type=FilterType.AND,
                filters=[allowed_filter, filter],
            )
    else:
        filter = allowed_filter

    return filter



from fastapi import APIRouter, Depends, HTTPException, status

from pytupli.server.api.dependencies import get_db_handler
from pytupli.schema import (
    Benchmark,
    BenchmarkHeader,
    BenchmarkQuery,
    UserOut,
    BaseFilter,
    RESOURCE_TYPE,
    RIGHT,
)
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.config import (  # environment variables, constants and Handler Factory
    BENCHMARK_COLLECTION_NAME,
)
from pytupli.server.management.authorization import (
    check_authorization,
    check_group_permissions,
    inject_read_permission_filter,
)

router = APIRouter()


@router.post('/create')
async def benchmarks_create(
    benchmark: BenchmarkQuery,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.BENCHMARK, RIGHT.BENCHMARK_CREATE)
    ),
) -> BenchmarkHeader:
    user, _ = auth_result
    query = {'hash': benchmark.hash}
    existing_benchmark = await db_handler.get_item(BENCHMARK_COLLECTION_NAME, query)

    if existing_benchmark:
        benchmark_entry = BenchmarkHeader(**existing_benchmark)
        # check if the benchmark already exists
        if benchmark_entry.created_by == user.username:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT, detail='Benchmark already exists'
            )
        elif 'global' in benchmark_entry.published_in:
            raise HTTPException(
                status_code=status.HTTP_409_CONFLICT,
                detail='Benchmark already exists in global group',
            )

    try:
        new_benchmark = Benchmark.create_new(
            created_by=user.username,
            **benchmark.model_dump(),
        )
        await db_handler.create_item(BENCHMARK_COLLECTION_NAME, new_benchmark.model_dump())
        # return the benchmark item that has just been created
        return BenchmarkHeader(**new_benchmark.model_dump())
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to create benchmark: {str(e)}',
        )


@router.get('/load')
async def benchmarks_load(
    benchmark_id: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, BenchmarkHeader | None] = Depends(
        check_authorization(RESOURCE_TYPE.BENCHMARK, RIGHT.BENCHMARK_READ)
    ),
) -> Benchmark:
    user, benchmark_header = auth_result
    try:
        # Check if the benchmark exists (resource was loaded by authorization)
        if not benchmark_header:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Benchmark not found')

        # Load the full benchmark data from the database
        benchmark_entry = await db_handler.get_item(BENCHMARK_COLLECTION_NAME, {'id': benchmark_id})
        if not benchmark_entry:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail='Benchmark not found')

        # Return the benchmark
        return Benchmark(**benchmark_entry)
    except HTTPException as e:
        raise e
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to load benchmark: {str(e)}',
        )


@router.post('/list')
async def benchmarks_list(
    filter: BaseFilter = BaseFilter(),
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(check_authorization(RESOURCE_TYPE.BENCHMARK)),
) -> list[BenchmarkHeader]:
    user, _ = auth_result
    filter = await inject_read_permission_filter(filter, user, RESOURCE_TYPE.BENCHMARK, db_handler)
    try:
        benchmarks = await db_handler.query_items(BENCHMARK_COLLECTION_NAME, filter)
        return [BenchmarkHeader(**benchmark) for benchmark in benchmarks]
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to list benchmarks: {str(e)}',
        )


@router.put('/publish')
async def benchmarks_publish(
    benchmark_id: str,
    publish_in: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, BenchmarkHeader | None] = Depends(
        check_authorization(RESOURCE_TYPE.BENCHMARK, RIGHT.BENCHMARK_READ, requires_ownership=True)
    ),
) -> None:
    user, benchmark = auth_result

    if not benchmark:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail='Benchmark does not exists'
        )

    # Check if user has CONTENT_CREATE rights in the specified group
    await check_group_permissions(db_handler, user, publish_in, [RIGHT.BENCHMARK_CREATE])

    try:
        # Create the update dictionary
        published_in = benchmark.published_in
        if publish_in in published_in:
            return

        published_in.append(publish_in)
        update = {'$set': {'published_in': published_in}}

        # update the benchmark in the db
        query = {'id': benchmark_id}
        await db_handler.update_item(BENCHMARK_COLLECTION_NAME, query, update)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to publish benchmark: {str(e)}',
        )


@router.put('/unpublish')
async def benchmarks_unpublish(
    benchmark_id: str,
    unpublish_from: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, BenchmarkHeader | None] = Depends(
        check_authorization(RESOURCE_TYPE.BENCHMARK, RIGHT.BENCHMARK_READ)
    ),
) -> None:
    user, benchmark = auth_result

    if not benchmark:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail='Benchmark does not exist'
        )

    # Check if user has BENCHMARK_DELETE rights in the specified group
    await check_group_permissions(db_handler, user, unpublish_from, [RIGHT.BENCHMARK_DELETE])

    # Check if currently published in this group
    current_published_in = benchmark.published_in
    if unpublish_from not in current_published_in:
        return

    # Remove the group from published_in list
    updated_published_in = [group for group in current_published_in if group != unpublish_from]
    update = {'$set': {'published_in': updated_published_in}}

    try:
        query = {'id': benchmark_id}
        await db_handler.update_item(BENCHMARK_COLLECTION_NAME, query, update)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to unpublish benchmark: {str(e)}',
        )


@router.delete('/delete')
async def benchmarks_delete(
    benchmark_id: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    _: tuple[UserOut, dict] = Depends(
        check_authorization(RESOURCE_TYPE.BENCHMARK, RIGHT.BENCHMARK_DELETE)
    ),
) -> None:
    try:
        # Delete the benchmark from the database (resource existence already verified by authorization)
        query = {'id': benchmark_id}
        await db_handler.delete_item(BENCHMARK_COLLECTION_NAME, query)
    except HTTPException as e:
        raise e
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to delete benchmark: {str(e)}',
        )



from fastapi import APIRouter

from pytupli.server.api.endpoints import artifacts, benchmarks, episodes, users, groups, roles

api_router = APIRouter()
api_router.include_router(artifacts.router, prefix='/artifacts', tags=['Artifacts'])
api_router.include_router(benchmarks.router, prefix='/benchmarks', tags=['Benchmarks'])
api_router.include_router(episodes.router, prefix='/episodes', tags=['Episodes'])
api_router.include_router(users.router, prefix='/access/users', tags=['Access', 'Users'])
api_router.include_router(groups.router, prefix='/access/groups', tags=['Access', 'Groups'])
api_router.include_router(roles.router, prefix='/access/roles', tags=['Access', 'Roles'])



from fastapi import APIRouter, Depends, HTTPException, status

from pytupli.server.api.dependencies import get_db_handler
from pytupli.schema import (
    BenchmarkHeader,
    EpisodeHeader,
    EpisodeItem,
    Episode,
    EpisodesListRequest,
    UserOut,
    RESOURCE_TYPE,
    RIGHT,
)
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.config import (  # environment variables, constants and Handler Factory
    BENCHMARK_COLLECTION_NAME,
    EPISODES_COLLECTION_NAME,
)
from pytupli.server.management.authorization import (
    check_authorization,
    check_group_permissions,
    inject_read_permission_filter,
    get_rights_in_groups,
)

router = APIRouter()


@router.post('/record')
async def episodes_record(
    episode: Episode,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(
        check_authorization(RESOURCE_TYPE.EPISODE, RIGHT.EPISODE_CREATE)
    ),
) -> EpisodeHeader:
    user, _ = auth_result

    # check if the benchmark exists
    query = {'id': episode.benchmark_id}
    benchmark_data = await db_handler.get_item(BENCHMARK_COLLECTION_NAME, query)
    if not benchmark_data:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Benchmark id doesn't exist"
        )

    benchmark = BenchmarkHeader(**benchmark_data)

    # check if user has read access to the benchmark
    user_rights = await get_rights_in_groups(db_handler, user, ['global', *benchmark.published_in])
    if RIGHT.BENCHMARK_READ not in user_rights:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail='Cannot record episodes for benchmarks without read access',
        )

    # record the episode
    try:
        episode_item = EpisodeItem.create_new(created_by=user.username, **episode.model_dump())
        await db_handler.create_item(EPISODES_COLLECTION_NAME, episode_item.model_dump())
        return episode_item
    except Exception as e:
        raise HTTPException(status_code=500, detail=f'Failed to record episode: {str(e)}')


@router.put('/publish')
async def episodes_publish(
    episode_id: str,
    publish_in: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, EpisodeItem | None] = Depends(
        check_authorization(RESOURCE_TYPE.EPISODE, RIGHT.EPISODE_READ, requires_ownership=True)
    ),
) -> None:
    user, episode = auth_result

    if not episode:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'Referenced episode {episode_id} does not exist',
        )

    # Check if user has CONTENT_CREATE rights in the specified group
    await check_group_permissions(db_handler, user, publish_in, [RIGHT.EPISODE_CREATE])

    # get referenced benchmark
    query = {'id': episode.benchmark_id}
    benchmark = BenchmarkHeader(**await db_handler.get_item(BENCHMARK_COLLECTION_NAME, query))

    # check if user has read access to the benchmark
    user_rights = await get_rights_in_groups(db_handler, user, ['global', *benchmark.published_in])
    if RIGHT.BENCHMARK_READ not in user_rights:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail=f'Insufficient permissions to read benchmark: {episode.benchmark_id}',
        )

    try:
        query = {'id': episode_id}
        published_in = episode.published_in or []
        if publish_in in published_in:
            return

        published_in.append(publish_in)
        update = {'$set': {'published_in': published_in}}
        await db_handler.update_items(EPISODES_COLLECTION_NAME, query, update)
    except Exception as e:
        raise HTTPException(status_code=500, detail=f'Failed to record episodes: {str(e)}')


@router.put('/unpublish')
async def episodes_unpublish(
    episode_id: str,
    unpublish_from: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, EpisodeItem | None] = Depends(
        check_authorization(RESOURCE_TYPE.EPISODE, RIGHT.EPISODE_READ)
    ),
) -> None:
    user, episode = auth_result

    if not episode:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f'Episode {episode_id} does not exist',
        )

    # Check if user has EPISODE_DELETE rights in the specified group
    await check_group_permissions(db_handler, user, unpublish_from, [RIGHT.EPISODE_DELETE])

    # Check if currently published in this group
    current_published_in = episode.published_in or []
    if unpublish_from not in current_published_in:
        return

    # Remove the group from published_in list
    updated_published_in = [group for group in current_published_in if group != unpublish_from]
    update = {'$set': {'published_in': updated_published_in}}

    try:
        query = {'id': episode_id}
        await db_handler.update_items(EPISODES_COLLECTION_NAME, query, update)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to unpublish episode: {str(e)}',
        )


@router.delete('/delete')
async def episodes_delete(
    episode_id: str,
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, dict] = Depends(
        check_authorization(RESOURCE_TYPE.EPISODE, RIGHT.EPISODE_DELETE)
    ),
) -> None:
    try:
        # Delete the episode(s) from the database (existence already verified by authorization)
        query = {'id': episode_id}
        await db_handler.delete_items(EPISODES_COLLECTION_NAME, query)
    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to delete episode(s): {str(e)}',
        )


@router.post('/list')
async def episodes_list(
    request: EpisodesListRequest = EpisodesListRequest(),
    db_handler: MongoDBHandler = Depends(get_db_handler),
    auth_result: tuple[UserOut, None] = Depends(check_authorization(RESOURCE_TYPE.EPISODE)),
) -> list[EpisodeHeader] | list[EpisodeItem]:
    user, _ = auth_result
    filter = await inject_read_permission_filter(request, user, RESOURCE_TYPE.EPISODE, db_handler)
    try:
        episodes = await db_handler.query_items(
            EPISODES_COLLECTION_NAME,
            filter,
            projection={'tuples': 0}
            if not request.include_tuples
            else None,  # Optionally exlude tuples from the result to reduce traffic
        )

        return (
            [EpisodeHeader(**episode) for episode in episodes]
            if not request.include_tuples
            else [EpisodeItem(**episode) for episode in episodes]
        )

    except Exception as e:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f'Failed to list episodes: {str(e)}',
        )



"""
Module for everything related to storage.
"""

from __future__ import annotations
import hashlib
import json
import logging
from pathlib import Path
from typing import Any, Callable, Dict

import pandas as pd
import requests
import keyring

from pytupli.schema import (
    ArtifactMetadata,
    ArtifactMetadataItem,
    BaseFilter,
    Benchmark,
    BenchmarkHeader,
    BenchmarkQuery,
    User,
    UserOut,
    UserRole,
    Episode,
    EpisodeHeader,
    EpisodeItem,
    FilterType,
    Group,
    GroupMembershipQuery,
    GroupWithMembers,
)

# Set up logger
logger = logging.getLogger(__name__)


class TupliStorageError(Exception):
    """
    Exception raised for errors in the storage operations.
    """

    pass


class TupliStorage:
    """
    Base class for storing StorableObjects.
    """

    def __init__(self):
        raise NotImplementedError

    def store_benchmark(self, benchmark_query: BenchmarkQuery) -> BenchmarkHeader:
        """
        Saves the serialized object to the specified storage.

        Args:
            benchmark_query (BenchmarkQuery): The serialized benchmark object to be saved.

        Returns:
            BenchmarkHeader: The header of the saved benchmark.
        """
        raise NotImplementedError

    def load_benchmark(
        self,
        uri: str,
    ) -> Benchmark:
        """
        Loads data from the specified URI.

        Args:
            uri (str): The URI of the data to be loaded.

        Returns:
            Benchmark: The loaded benchmark object.
        """
        raise NotImplementedError

    def list_benchmarks(self, filter: BaseFilter) -> list[BenchmarkHeader]:
        """
        Lists all benchmarks in the storage that match the specified filter.

        Args:
            filter (BaseFilter): The filter to apply when listing benchmarks.

        Returns:
            list[BenchmarkHeader]: A list of benchmark headers that match the filter.
        """
        raise NotImplementedError

    def delete_benchmark(self, uri: str) -> None:
        """
        Deletes the specified benchmark from the storage.
        Args:
            uri (str): The URI/ID of the benchmark to delete.
        """
        raise NotImplementedError

    def store_artifact(self, artifact: bytes, metadata: ArtifactMetadata) -> ArtifactMetadataItem:
        """
        Stores the artifact in the storage.

        Args:
            artifact (bytes): The artifact to store.
            metadata (ArtifactMetadata): Metadata for the artifact.

        Returns:
            ArtifactMetadataItem: Metadata item for the stored artifact.
        """
        raise NotImplementedError

    def load_artifact(self, uri: str, **kwargs) -> pd.DataFrame:
        """
        Loads the artifact from the storage.

        Args:
            uri (str): The URI/ID of the artifact to load.
            kwargs: Additional arguments for loading the artifact.
        """
        raise NotImplementedError

    def list_artifacts(self, filter: BaseFilter) -> list[ArtifactMetadataItem]:
        """
        Lists all artifacts in the storage that match the specified filter.

        Args:
            filter (BaseFilter): The filter to apply when listing artifacts.

        Returns:
            list[ArtifactMetadataItem]: A list of artifacts that match the filter.
        """
        raise NotImplementedError

    def delete_artifact(self, uri: str) -> None:
        """
        Deletes the specified artifact from the storage.

        Args:
            uri (str): The URI/ID of the artifact to delete.
        """
        raise NotImplementedError

    def record_episode(self, episode: Episode) -> EpisodeHeader:
        """
        Records an episode in the storage.

        Args:
            episode (Episode): The episode to record.

        Returns:
            EpisodeHeader: The header of the recorded episode.
        """
        raise NotImplementedError

    def publish_episode(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishes the specified episode in the storage.

        Args:
            uri (str): The URI/ID of the episode to publish.
            publish_in (str): The group to publish the episode in. Defaults to 'global'.
        """
        raise NotImplementedError

    def list_episodes(
        self, filter: BaseFilter = None, include_tuples: bool = False
    ) -> list[EpisodeHeader] | list[EpisodeItem]:
        """
        Lists all episodes in the storage that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing episodes.
            include_tuples (bool, optional): Whether to include tuples in the episode data.

        Returns:
            list[EpisodeHeader] | list[EpisodeItem]: A list of episode headers or full episode items with tuples.
        """
        raise NotImplementedError

    def delete_episode(self, uri: str) -> None:
        """
        Deletes the specified episode from the storage.

        Args:
            uri (str): The URI/ID of the episode to delete.
        """
        raise NotImplementedError


class TupliAPIClient(TupliStorage):
    """
    Class for storing StorableObjects in the API.

    This class provides methods for interacting with the Tupli API, including user management,
    group management, role management, benchmark operations, artifact handling, and episode management.

    Methods:
        User Management:
            signup(username: str, password: str) -> User
                Creates a new user account.
            login(username: str, password: str, url: str | None = None) -> None
                Authenticates with the API and stores access tokens.
            list_users() -> list[User]
                Lists all users.
            change_password(username: str, new_password: str) -> None
                Changes a user's password.
            delete_user(username: str) -> None
                Deletes a user and their content.

        Group Management:
            create_group(group: Group) -> Group
                Creates a new group.
            list_groups() -> list[Group]
                Lists all groups accessible to the current user.
            read_group(group_name: str) -> GroupWithMembers
                Reads a specific group with its members.
            delete_group(group_name: str) -> None
                Deletes a group.
            add_group_members(group_membership_query: GroupMembershipQuery) -> GroupWithMembers
                Adds members to a group with specified roles.
            remove_group_members(group_membership_query: GroupMembershipQuery) -> GroupWithMembers
                Removes members from a group.

        Role Management:
            list_roles() -> list[UserRole]
                Lists all available user roles.
            create_role(role: UserRole) -> UserRole
                Creates a new role.
            delete_role(role_name: str) -> None
                Deletes a role.

        Benchmark Operations:
            store_benchmark(benchmark_query: BenchmarkQuery) -> BenchmarkHeader
                Saves a benchmark to the API.
            load_benchmark(uri: str) -> Benchmark
                Loads a benchmark from the API.
            list_benchmarks(filter: BaseFilter = None) -> list[BenchmarkHeader]
                Lists benchmarks matching the filter.
            delete_benchmark(uri: str) -> None
                Deletes a benchmark.
            publish_benchmark(uri: str) -> None
                Publishes a benchmark.

        Artifact Operations:
            store_artifact(artifact: bytes, metadata: ArtifactMetadata) -> ArtifactMetadataItem
                Stores an artifact in the API.
            load_artifact(uri: str, **kwargs) -> bytes
                Loads an artifact from the API.
            list_artifacts(filter: BaseFilter = None) -> list[ArtifactMetadataItem]
                Lists artifacts matching the filter.
            delete_artifact(uri: str) -> None
                Deletes an artifact.
            publish_artifact(uri: str) -> None
                Publishes an artifact.

        Episode Operations:
            record_episode(episode: Episode) -> EpisodeHeader
                Records an episode in the API.
            publish_episode(uri: str) -> None
                Publishes an episode.
            list_episodes(filter: BaseFilter = None, include_tuples: bool = False) -> list[EpisodeHeader] | list[EpisodeItem]
                Lists episodes matching the filter.
            delete_episode(uri: str) -> None
                Deletes an episode.

        Configuration:
            set_url(url: str) -> None
                Sets the base URL for the API.
    """

    def __init__(self) -> TupliAPIClient:
        self.base_url = keyring.get_password('pytupli', 'base_url')

        if not self.base_url:
            self.base_url = 'http://localhost:8080'

    def _get_bearer_token(self) -> dict:
        """
        Gets the bearer token for API requests.
        First tries to use the stored access token, and if that fails,
        tries to refresh the token.

        Returns:
            dict: Headers with the bearer token.
        """
        # Try to get the stored access token
        access_token = keyring.get_password('pytupli', 'access_token')

        if access_token:
            # First try to use the existing token
            return {'Authorization': f'Bearer {access_token}'}

        # If no access token stored, refresh token
        return self._refresh_token()

    def _refresh_token(self) -> dict:
        """
        Refreshes the access token using the stored refresh token.

        Returns:
            dict: Headers with the refreshed bearer token.

        Raises:
            TupliStorageError: If the refresh token is not available or the refresh fails.
        """
        refresh_token = keyring.get_password('pytupli', 'refresh_token')

        if not refresh_token:
            raise TupliStorageError('No refresh token available. Please login first.')

        try:
            response = requests.post(
                f'{self.base_url}/access/users/refresh-token',
                headers={'Authorization': f'Bearer {refresh_token}'},
            )
            response.raise_for_status()

            new_access_token = response.json()['access_token']
            # Store the new token
            keyring.set_password('pytupli', 'access_token', new_access_token)

            return {'Authorization': f'Bearer {new_access_token}'}
        except Exception as e:
            # If refresh fails, both tokens might be invalid
            keyring.delete_password('pytupli', 'access_token')
            keyring.delete_password('pytupli', 'refresh_token')
            raise TupliStorageError(f'Token refresh failed: {str(e)}. Please login again.')

    def _authenticated_request(self, method, url, **kwargs) -> requests.Response:
        """
        Executes an authenticated request to the API.
        Handles token refresh if the access token is expired.

        Args:
            method (str): HTTP method (get, post, put, delete)
            url (str): URL for the request
            **kwargs: Additional arguments for the request

        Returns:
            Response: The response from the request

        Raises:
            TupliStorageError: If the request fails or the token refresh fails.
        """
        # First try with current access token
        try:
            headers = self._get_bearer_token()
            if 'headers' in kwargs:
                kwargs['headers'].update(headers)
            else:
                kwargs['headers'] = headers

            response = getattr(requests, method.lower())(url, **kwargs)
            response.raise_for_status()
            return response
        except requests.HTTPError as e:
            if e.response.status_code == 401:  # Unauthorized, token might be expired
                # Try to refresh token and retry
                headers = self._refresh_token()
                if 'headers' in kwargs:
                    kwargs['headers'].update(headers)
                else:
                    kwargs['headers'] = headers

                response = getattr(requests, method.lower())(url, **kwargs)
                response.raise_for_status()
                return response
            raise TupliStorageError(f'API request failed: {str(e)}')
        except Exception as e:
            raise TupliStorageError(f'Request failed: {str(e)}')

    def set_url(self, url: str) -> None:
        """
        Sets the base URL for the API.

        Args:
            url (str): The base URL for the API.
        """
        self.base_url = url
        keyring.set_password('pytupli', 'base_url', url)

    # User management methods
    def signup(self, username: str, password: str) -> User:
        """
        Creates a new user account.

        Args:
            username (str): The username for the new account
            password (str): The password for the new account

        Returns:
            User: The created user object
        """
        try:  # first try authenticated request
            response = self._authenticated_request(
                'post',
                f'{self.base_url}/access/users/create',
                json={'username': username, 'password': password},
            )
            return UserOut(**response.json())
        except TupliStorageError:  # if that fails, try unauthenticated request
            try:
                response = requests.post(
                    f'{self.base_url}/access/users/create',
                    json={'username': username, 'password': password},
                )
                response.raise_for_status()
                return UserOut(**response.json())
            except requests.HTTPError as e:
                raise TupliStorageError(f'Signup failed: {str(e)}')

    def login(self, username: str, password: str, url: str | None = None) -> None:
        """
        Authenticates with the API and stores the access and refresh tokens.

        Args:
            username (str): The username for the API.
            password (str): The password for the API.
            url (str, optional): The base URL for the API.
                If not provided, uses the stored URL.
                If provided, it will be set as the new base URL.
        """

        if url:
            self.set_url(url)

        try:
            response = requests.post(
                f'{self.base_url}/access/users/token',
                json={'username': username, 'password': password},
            )
            response.raise_for_status()
        except requests.HTTPError as e:
            raise TupliStorageError(f'Login failed: {str(e)}')

        data = response.json()
        access_token = data['access_token']['token']
        refresh_token = data['refresh_token']['token']

        # Store tokens in keyring
        keyring.set_password('pytupli', 'access_token', access_token)
        keyring.set_password('pytupli', 'refresh_token', refresh_token)

    def list_users(self) -> list[User]:
        """
        Lists all users.

        Returns:
            list[User]: A list of all users
        """
        response = self._authenticated_request('get', f'{self.base_url}/access/users/list')
        return [UserOut(**user) for user in response.json()]

    def list_roles(self) -> list[UserRole]:
        """
        Lists all user roles.

        Returns:
            list[UserRole]: A list of all user roles
        """
        response = self._authenticated_request('get', f'{self.base_url}/access/roles/list')
        return [UserRole(**role) for role in response.json()]

    def change_password(self, username: str, new_password: str) -> None:
        """
        Changes a user's password.

        Args:
            username (str): The username of the account to change
            new_password (str): The new password
        """
        _ = self._authenticated_request(
            'put',
            f'{self.base_url}/access/users/change-password',
            json={'username': username, 'password': new_password},
        )

    def delete_user(self, username: str) -> None:
        """
        Deletes a user and all their content.

        Args:
            username (str): The username of the account to delete
        """
        self._authenticated_request(
            'delete', f'{self.base_url}/access/users/delete', params={'username': username}
        )

    # Group management methods
    def create_group(self, group: Group) -> Group:
        """
        Creates a new group.

        Args:
            group (Group): The group to create

        Returns:
            Group: The created group object
        """
        response = self._authenticated_request(
            'post',
            f'{self.base_url}/access/groups/create',
            json=group.model_dump(),
        )
        return Group(**response.json())

    def list_groups(self) -> list[Group]:
        """
        Lists all groups accessible to the current user.

        Returns:
            list[Group]: A list of groups
        """
        response = self._authenticated_request('get', f'{self.base_url}/access/groups/list')
        return [Group(**group) for group in response.json()]

    def read_group(self, group_name: str) -> GroupWithMembers:
        """
        Reads a specific group with its members.

        Args:
            group_name (str): The name of the group to read

        Returns:
            GroupWithMembers: The group with its members
        """
        response = self._authenticated_request(
            'get', f'{self.base_url}/access/groups/read', params={'group_name': group_name}
        )
        return GroupWithMembers(**response.json())

    def delete_group(self, group_name: str) -> None:
        """
        Deletes a group.

        Args:
            group_name (str): The name of the group to delete
        """
        self._authenticated_request(
            'delete', f'{self.base_url}/access/groups/delete', params={'group_name': group_name}
        )

    def add_group_members(self, group_membership_query: GroupMembershipQuery) -> GroupWithMembers:
        """
        Adds members to a group with specified roles.

        Args:
            group_membership_query (GroupMembershipQuery): The membership query specifying group and members

        Returns:
            GroupWithMembers: The updated group with members
        """
        response = self._authenticated_request(
            'post',
            f'{self.base_url}/access/groups/add-members',
            json=group_membership_query.model_dump(),
        )
        return GroupWithMembers(**response.json())

    def remove_group_members(
        self, group_membership_query: GroupMembershipQuery
    ) -> GroupWithMembers:
        """
        Removes members from a group.

        Args:
            group_membership_query (GroupMembershipQuery): The membership query specifying group and members

        Returns:
            GroupWithMembers: The updated group with members
        """
        response = self._authenticated_request(
            'post',
            f'{self.base_url}/access/groups/remove-members',
            json=group_membership_query.model_dump(),
        )
        return GroupWithMembers(**response.json())

    # Role management methods
    def create_role(self, role: UserRole) -> UserRole:
        """
        Creates a new role.

        Args:
            role (UserRole): The role to create

        Returns:
            UserRole: The created role object
        """
        response = self._authenticated_request(
            'post',
            f'{self.base_url}/access/roles/create',
            json=role.model_dump(),
        )
        return UserRole(**response.json())

    def delete_role(self, role_name: str) -> None:
        """
        Deletes a role.

        Args:
            role_name (str): The name of the role to delete
        """
        self._authenticated_request(
            'delete', f'{self.base_url}/access/roles/delete', params={'role_name': role_name}
        )

    def store_benchmark(self, benchmark_query: BenchmarkQuery) -> BenchmarkHeader:
        """
        Saves the serialized object to the API.

        Args:
            benchmark_query (BenchmarkQuery): The serialized benchmark object to be saved as well as some metadata.

        Returns:
            BenchmarkHeader: The header of the saved benchmark.
        """
        response = self._authenticated_request(
            'post', f'{self.base_url}/benchmarks/create', json=benchmark_query.model_dump()
        )
        return BenchmarkHeader(**response.json())

    def load_benchmark(self, uri: str) -> Benchmark:
        """
        Loads the serialized benchmark from the API.

        Args:
            uri (str): hash of the object to be loaded.

        Returns:
            Benchmark: The loaded benchmark object.
        """
        response = self._authenticated_request(
            'get', f'{self.base_url}/benchmarks/load?benchmark_id={uri}'
        )
        return Benchmark(**response.json())

    def store_artifact(self, artifact: bytes, metadata: ArtifactMetadata) -> ArtifactMetadataItem:
        """
        Stores the artifact in the API.

        Args:
            artifact (bytes): The artifact to store.
            metadata (dict, optional): Metadata for the artifact.

        Returns:
            ArtifactMetadataItem: Metadata item for the stored artifact.
        """
        response = self._authenticated_request(
            'post',
            f'{self.base_url}/artifacts/upload',
            files={'data': artifact},
            data={'metadata': metadata.model_dump_json(serialize_as_any=True)},
        )
        return ArtifactMetadataItem(**response.json())

    def load_artifact(self, uri: str, **kwargs) -> bytes:
        """
        Load artifact from the API.

        Args:
            uri (str): hash of the object to be loaded.

        Returns:
            bytes: The raw artifact data
        """
        response = self._authenticated_request(
            'get', f'{self.base_url}/artifacts/download?artifact_id={uri}'
        )
        return response.content

    def publish_benchmark(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishes the benchmark in the API.

        Args:
            uri (str): The hash of the benchmark to be published.
            publish_in (str): The group to publish the benchmark in. Defaults to 'global'.
        """
        self._authenticated_request(
            'put', f'{self.base_url}/benchmarks/publish?benchmark_id={uri}&publish_in={publish_in}'
        )

    def delete_benchmark(self, uri: str) -> None:
        """
        Deletes the specified object from the API.

        Args:
            uri (str): The hash of the object to be deleted.
        """
        self._authenticated_request(
            'delete', f'{self.base_url}/benchmarks/delete?benchmark_id={uri}'
        )

    def delete_artifact(self, uri: str) -> None:
        """
        Deletes the specified artifact from the API.

        Args:
            uri (str): The hash of the artifact to be deleted.
        """
        self._authenticated_request('delete', f'{self.base_url}/artifacts/delete?artifact_id={uri}')

    def publish_artifact(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishes the artifact in the API.

        Args:
            uri (str): The hash of the artifact to be published.
            publish_in (str): The group to publish the artifact in. Defaults to 'global'.
        """
        self._authenticated_request(
            'put', f'{self.base_url}/artifacts/publish?artifact_id={uri}&publish_in={publish_in}'
        )

    def list_benchmarks(self, filter: BaseFilter = None) -> list[BenchmarkHeader]:
        """
        Lists all benchmarks in the storage that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing benchmarks.

        Returns:
            list[BenchmarkHeader]: A list of benchmark headers that match the filter.
        """
        json_data = filter.model_dump() if filter else {}

        response = self._authenticated_request(
            'post', f'{self.base_url}/benchmarks/list', json=json_data
        )
        return [BenchmarkHeader(**benchmark) for benchmark in response.json()]

    def list_artifacts(self, filter: BaseFilter = None) -> list[ArtifactMetadataItem]:
        """
        Lists all artifacts in the storage that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing artifacts.

        Returns:
            list[ArtifactMetadataItem]: A list of artifacts that match the filter.
        """
        json_data = filter.model_dump() if filter else {}

        response = self._authenticated_request(
            'post', f'{self.base_url}/artifacts/list', json=json_data
        )
        return [ArtifactMetadataItem(**artifact) for artifact in response.json()]

    # Episode-related methods
    def record_episode(self, episode: Episode) -> EpisodeHeader:
        """
        Records an episode in the API.

        Args:
            episode (Episode): The episode to record.

        Returns:
            EpisodeHeader: The header of the recorded episode.
        """
        response = self._authenticated_request(
            'post',
            f'{self.base_url}/episodes/record',
            json=episode.model_dump(),
        )
        episode_data = response.json()
        return EpisodeHeader(**episode_data)

    def publish_episode(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishes an episode in the API.

        Args:
            uri (str): The ID of the episode to publish.
            publish_in (str): The group to publish the episode in. Defaults to 'global'.
        """
        self._authenticated_request(
            'put',
            f'{self.base_url}/episodes/publish?episode_id={uri}&publish_in={publish_in}',
        )

    def list_episodes(
        self, filter: BaseFilter = None, include_tuples: bool = False
    ) -> list[EpisodeHeader] | list[EpisodeItem]:
        """
        Lists all episodes in the API that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing episodes.
            include_tuples (bool, optional): Whether to include tuples in the episode data.

        Returns:
            list[EpisodeHeader] | list[EpisodeItem]: A list of episode headers or full episode items.
        """
        json_data = {
            **(filter.model_dump() if filter else {}),
            'include_tuples': include_tuples,
        }

        response = self._authenticated_request(
            'post',
            f'{self.base_url}/episodes/list',
            json=json_data,
        )

        if include_tuples:
            return [EpisodeItem(**episode) for episode in response.json()]
        else:
            return [EpisodeHeader(**episode) for episode in response.json()]

    def delete_episode(self, uri: str) -> None:
        """
        Deletes an episode from the API.

        Args:
            uri (str): The ID of the episode to delete.
        """
        self._authenticated_request(
            'delete',
            f'{self.base_url}/episodes/delete?episode_id={uri}',
        )

    def unpublish_benchmark(self, uri: str, unpublish_from: str) -> None:
        """
        Unpublishes a benchmark from the specified group.

        Args:
            uri (str): The ID of the benchmark to unpublish.
            unpublish_from (str): The group to unpublish the benchmark from.
        """
        self._authenticated_request(
            'put',
            f'{self.base_url}/benchmarks/unpublish?benchmark_id={uri}&unpublish_from={unpublish_from}',
        )

    def unpublish_artifact(self, uri: str, unpublish_from: str) -> None:
        """
        Unpublishes an artifact from the specified group.

        Args:
            uri (str): The ID of the artifact to unpublish.
            unpublish_from (str): The group to unpublish the artifact from.
        """
        self._authenticated_request(
            'put',
            f'{self.base_url}/artifacts/unpublish?artifact_id={uri}&unpublish_from={unpublish_from}',
        )

    def unpublish_episode(self, uri: str, unpublish_from: str) -> None:
        """
        Unpublishes an episode from the specified group.

        Args:
            uri (str): The ID of the episode to unpublish.
            unpublish_from (str): The group to unpublish the episode from.
        """
        self._authenticated_request(
            'put',
            f'{self.base_url}/episodes/unpublish?episode_id={uri}&unpublish_from={unpublish_from}',
        )


class FileStorage(TupliStorage):
    """
    Storage class for saving and loading benchmarks to/from files.
    """

    def __init__(
        self,
        storage_base_dir: str = '_local_storage',
    ) -> FileStorage:
        """
        Initializes the FileStorage with the specified base directory.

        Args:
            storage_base_dir (str): The base directory for storage.

        Returns:
            FileStorage: The initialized FileStorage object.

        Raises:
            TupliStorageError: If the base directory cannot be created.
        """
        self.storage_dir = Path(storage_base_dir)
        # Create base storage directory if it doesn't exist
        try:
            self.storage_dir.mkdir(exist_ok=True)
        except Exception as e:
            raise TupliStorageError(f'Failed to create storage directory: {str(e)}')

    def store_benchmark(self, benchmark_query: BenchmarkQuery) -> BenchmarkHeader:
        """
        Saves the benchmark object to a file.

        Args:
            benchmark_query (BenchmarkQuery): The benchmark query to be saved.

        Returns:
            BenchmarkHeader: The header of the saved benchmark.

        Raises:
            TupliStorageError: If the benchmark cannot be saved or serialized.
        """
        # Create a directory if it doesn't exist
        directory = self.storage_dir / 'benchmarks'
        try:
            directory.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            raise TupliStorageError(f'Failed to create benchmarks directory: {str(e)}')

        # Create a benchmark from the query
        benchmark = Benchmark.create_new(**benchmark_query.model_dump(), created_by='local_storage')

        # Check if benchmark with the same ID already exists
        for existing_file in directory.glob('*.json'):
            try:
                with open(existing_file, 'r', encoding='UTF-8') as f:
                    existing_benchmark = json.loads(f.read())
                if existing_benchmark['hash'] == benchmark.hash:
                    logger.info('Benchmark with hash %s already exists', benchmark.hash)
                    return BenchmarkHeader(**existing_benchmark)
            except Exception as e:
                logger.warning(f'Error reading benchmark file {existing_file}: {str(e)}')
                continue

        # Create filename based on benchmark ID
        file_name = f'{benchmark.id}.json'
        file_path = directory / file_name

        if file_path.exists():
            raise TupliStorageError(
                f'The file {file_path} already exists and will not be overwritten.'
            )

        # Serialize the benchmark to JSON
        try:
            serialized_object = json.dumps(benchmark.model_dump(), indent=2)
        except Exception as e:
            raise TupliStorageError(f'Failed to serialize benchmark: {str(e)}')

        try:
            with open(file_path, 'w', encoding='UTF-8') as f:
                f.write(serialized_object)

            # Check if the file was saved correctly
            if not file_path.exists():
                raise TupliStorageError(f'Failed to save benchmark to {file_path}')
            else:
                logger.info('Saved benchmark to %s', file_path)
        except Exception as e:
            raise TupliStorageError(f'Failed to write benchmark to file: {str(e)}')

        # Return the benchmark header
        return BenchmarkHeader(**benchmark.model_dump())

    def load_benchmark(self, uri: str) -> Benchmark:
        """
        Loads a benchmark from the file using the benchmark ID.

        Args:
            uri (str): The ID of the benchmark to be loaded.

        Returns:
            Benchmark: The loaded benchmark object.

        Raises:
            TupliStorageError: If the benchmark cannot be loaded or parsed.
        """
        # Construct the file path from the benchmark ID
        file_path = self.storage_dir / 'benchmarks' / f'{uri}.json'

        if not file_path.exists():
            raise TupliStorageError(f'Benchmark with ID {uri} does not exist.')

        try:
            with open(file_path, 'r', encoding='UTF-8') as f:
                benchmark_dict = json.loads(f.read())
        except json.JSONDecodeError as e:
            raise TupliStorageError(f'Failed to parse JSON from benchmark {uri}: {str(e)}')
        except Exception as e:
            raise TupliStorageError(f'Failed to read benchmark file for {uri}: {str(e)}')

        try:
            # Create and return a Benchmark object from the loaded JSON
            return Benchmark(**benchmark_dict)
        except Exception as e:
            raise TupliStorageError(f'Invalid benchmark data for {uri}: {str(e)}')

    # Helper methods for artifacts
    def store_artifact(self, artifact: bytes, metadata: ArtifactMetadata) -> ArtifactMetadataItem:
        """
        Stores an artifact as a file and returns its metadata.

        Args:
            artifact (bytes): The artifact data to store.
            metadata (ArtifactMetadata): Metadata for the artifact.

        Returns:
            ArtifactMetadataItem: Metadata for the stored artifact.

        Raises:
            TupliStorageError: If the artifact cannot be stored or metadata cannot be serialized.
        """
        # Create a directory for artifacts if it doesn't exist
        data_dir = self.storage_dir / 'artifacts'
        try:
            data_dir.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            raise TupliStorageError(f'Failed to create artifacts directory: {str(e)}')

        # Generate hash for the artifact
        artifact_hash = hashlib.sha256(artifact).hexdigest()

        # Check if artifact with the same hash already exists
        for metadata_path in data_dir.glob('*.metadata.json'):
            try:
                with open(metadata_path, 'r', encoding='UTF-8') as f:
                    existing_metadata = json.loads(f.read())
                if existing_metadata.get('hash') == artifact_hash:
                    logger.info('Artifact with hash %s already exists', artifact_hash)
                    return ArtifactMetadataItem(**existing_metadata)
            except Exception as e:
                logger.warning(f'Error reading metadata file {metadata_path}: {str(e)}')
                continue

        metadata_item = ArtifactMetadataItem.create_new(
            hash=artifact_hash,
            created_by='local_storage',
            **metadata.model_dump(),
        )

        # Create file path using the artifact ID
        file_path = data_dir / f'{metadata_item.id}'

        if file_path.exists():
            raise TupliStorageError(
                f'The file {file_path} already exists and will not be overwritten.'
            )

        try:
            # Write the artifact data to the file
            with open(file_path, 'wb') as f:
                f.write(artifact)

            # Store the metadata
            metadata_path = file_path.with_suffix('.metadata.json')

            with open(metadata_path, 'w', encoding='UTF-8') as f:
                json.dump(metadata_item.model_dump(serialize_as_any=True), f, indent=2)

            logger.info('Stored artifact to %s with metadata', file_path)
            return metadata_item
        except Exception as e:
            raise TupliStorageError(f'Failed to store artifact: {str(e)}')

    def load_artifact(self, uri: str, **kwargs) -> bytes:
        """
        Loads an artifact from a file.

        Args:
            uri (str): The ID of the artifact to load.
            kwargs: Additional arguments (ignored in file storage)

        Returns:
            bytes: The raw artifact data

        Raises:
            TupliStorageError: If the artifact cannot be loaded.
        """
        # Construct file path from artifact ID
        file_path = self.storage_dir / 'artifacts' / uri

        if not file_path.exists():
            raise TupliStorageError(f'Artifact with ID {uri} does not exist.')

        try:
            # Read the file as bytes
            with open(file_path, 'rb') as f:
                data = f.read()

            return data
        except Exception as e:
            raise TupliStorageError(f'Failed to load artifact {uri}: {str(e)}')

    def convert_filter_to_function(
        self, filter_obj: BaseFilter
    ) -> Callable[[Dict[str, Any]], bool]:
        """
        Convert a BaseFilter object to a filter function that can be applied to dictionaries.
        Supports nested dictionary access with keys in the form of "a.b.key".

        Args:
            filter_obj (BaseFilter): The filter object to convert.

        Returns:
            Callable[[Dict[str, Any]], bool]: A function that takes a dictionary and returns True if the dictionary matches the filter.

        Raises:
            TupliStorageError: If the filter type is unknown.
        """
        if filter_obj is None:
            return lambda item: True

        def get_nested_value(item: Dict[str, Any], key_path: str) -> Any:
            """Get value from nested dictionary using dot notation."""
            keys = key_path.split('.')
            value = item

            for k in keys:
                if not isinstance(value, dict) or k not in value:
                    return None
                value = value[k]

            return value

        def key_exists(item: Dict[str, Any], key_path: str) -> bool:
            """Check if a key path exists in nested dictionary."""
            keys = key_path.split('.')
            value = item

            for k in keys:
                if not isinstance(value, dict) or k not in value:
                    return False
                value = value[k]

            return True

        match filter_obj.type:
            case FilterType.AND:
                sub_filters = [self.convert_filter_to_function(f) for f in filter_obj.filters]
                return lambda item: all(f(item) for f in sub_filters)

            case FilterType.OR:
                sub_filters = [self.convert_filter_to_function(f) for f in filter_obj.filters]
                return lambda item: any(f(item) for f in sub_filters)

            case FilterType.EQ:
                return (
                    lambda item: key_exists(item, filter_obj.key)
                    and get_nested_value(item, filter_obj.key) == filter_obj.value
                )

            case FilterType.GEQ:
                return (
                    lambda item: key_exists(item, filter_obj.key)
                    and get_nested_value(item, filter_obj.key) >= filter_obj.value
                )

            case FilterType.LEQ:
                return (
                    lambda item: key_exists(item, filter_obj.key)
                    and get_nested_value(item, filter_obj.key) <= filter_obj.value
                )

            case FilterType.GT:
                return (
                    lambda item: key_exists(item, filter_obj.key)
                    and get_nested_value(item, filter_obj.key) > filter_obj.value
                )

            case FilterType.LT:
                return (
                    lambda item: key_exists(item, filter_obj.key)
                    and get_nested_value(item, filter_obj.key) < filter_obj.value
                )

            case FilterType.NE:
                return (
                    lambda item: key_exists(item, filter_obj.key)
                    and get_nested_value(item, filter_obj.key) != filter_obj.value
                )

            case _:
                raise TupliStorageError(f'Unknown filter type: {filter_obj.type}')

    def list_benchmarks(self, filter: BaseFilter = None) -> list[BenchmarkHeader]:
        """
        Lists all benchmarks in the storage that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing benchmarks.

        Returns:
            list[BenchmarkHeader]: A list of benchmark headers that match the filter.
        """
        benchmark_dir = self.storage_dir / 'benchmarks'
        if not benchmark_dir.exists():
            return []

        results = []
        filter_func = self.convert_filter_to_function(filter)

        for file_path in benchmark_dir.glob('*.json'):
            try:
                with open(file_path, 'r', encoding='UTF-8') as f:
                    benchmark_dict = json.loads(f.read())

                # Apply filter function
                if filter_func(benchmark_dict):
                    # Create header from benchmark data
                    header = BenchmarkHeader(**benchmark_dict)
                    results.append(header)

            except Exception as e:
                logger.info('Error loading benchmark header from %s: %s', file_path, str(e))
                continue

        return results

    def list_artifacts(self, filter: BaseFilter = None) -> list[ArtifactMetadataItem]:
        """
        Lists all artifacts in the storage that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing artifacts.

        Returns:
            list[ArtifactMetadataItem]: A list of artifacts that match the filter.
        """
        artifacts_dir = self.storage_dir / 'artifacts'
        if not artifacts_dir.exists():
            return []

        results = []
        filter_func = self.convert_filter_to_function(filter)

        # Look for files that have an accompanying metadata file
        for metadata_path in artifacts_dir.glob('*.metadata.json'):
            try:
                # Read metadata from JSON file
                with open(metadata_path, 'r', encoding='UTF-8') as f:
                    metadata_dict = json.loads(f.read())

                # Apply filter
                if filter_func(metadata_dict):
                    metadata_item = ArtifactMetadataItem(**metadata_dict)
                    results.append(metadata_item)

            except Exception as e:
                logger.info('Error loading artifact metadata from %s: %s', metadata_path, str(e))
                continue

        return results

    def delete_benchmark(self, uri: str) -> None:
        """
        Deletes the specified benchmark from the storage.

        Args:
            uri (str): The ID of the benchmark to delete.
        """
        # Construct file path from benchmark ID
        file_path = self.storage_dir / 'benchmarks' / f'{uri}.json'

        if not file_path.exists():
            raise TupliStorageError(f'Benchmark with ID {uri} does not exist.')

        try:
            file_path.unlink()
            logger.info('Deleted benchmark with ID %s', uri)
        except Exception as e:
            raise TupliStorageError(f'Failed to delete benchmark {uri}: {str(e)}')

    def delete_artifact(self, uri: str) -> None:
        """
        Deletes the specified artifact from the storage.

        Args:
            uri (str): The ID of the artifact to delete.
        """
        # Construct file path from artifact ID
        file_path = self.storage_dir / 'artifacts' / uri
        metadata_path = file_path.with_suffix('.metadata.json')

        if not file_path.exists():
            raise TupliStorageError(f'Artifact with ID {uri} does not exist.')

        try:
            file_path.unlink()
            if metadata_path.exists():
                metadata_path.unlink()
            logger.info('Deleted artifact with ID %s', uri)
        except Exception as e:
            raise TupliStorageError(f'Failed to delete artifact {uri}: {str(e)}')

    def publish_benchmark(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishing functionality is not available in FileStorage.
        This is a placeholder to implement the interface.

        Args:
            uri (str): The URI of the benchmark to publish.
            publish_in (str): The group to publish the benchmark in. Defaults to 'global'.

        Returns:
            str: The URI of the benchmark.
        """
        logger.info('Publishing functionality is not available in FileStorage')

    def publish_artifact(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishing functionality is not available in FileStorage.
        This is a placeholder to implement the interface.

        Args:
            uri (str): The URI of the artifact to publish.
            publish_in (str): The group to publish the artifact in. Defaults to 'global'.
        """
        logger.info('Publishing functionality is not available in FileStorage')

    # Episode-related methods
    def record_episode(self, episode: Episode) -> EpisodeHeader:
        """
        Records an episode in the local file storage.

        Args:
            episode (Episode): The episode to record.

        Returns:
            EpisodeHeader: The header of the recorded episode.

        Raises:
            TupliStorageError: If the episode cannot be saved or serialized.
        """
        # Create a directory for episodes if it doesn't exist
        episodes_dir = self.storage_dir / 'episodes'
        try:
            episodes_dir.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            raise TupliStorageError(f'Failed to create episodes directory: {str(e)}')

        # Check if the referenced benchmark exists
        benchmark_dir = self.storage_dir / 'benchmarks'
        benchmark_found = False
        for file_path in benchmark_dir.glob('*.json'):
            try:
                with open(file_path, 'r', encoding='UTF-8') as f:
                    benchmark_dict = json.loads(f.read())
                if benchmark_dict['id'] == episode.benchmark_id:
                    benchmark_found = True
                    break
            except Exception:
                continue

        if not benchmark_found:
            raise TupliStorageError(
                f'Referenced benchmark with ID {episode.benchmark_id} does not exist'
            )

        # Create a full EpisodeItem with metadata
        episode_item = EpisodeItem.create_new(**episode.model_dump(), created_by='local_storage')

        # Create filename based on episode ID
        file_name = f'{episode_item.id}.json'
        file_path = episodes_dir / file_name

        if file_path.exists():
            raise TupliStorageError(
                f'The file {file_path} already exists and will not be overwritten.'
            )

        # Serialize the episode to JSON
        try:
            serialized_object = json.dumps(episode_item.model_dump(), indent=2)
        except Exception as e:
            raise TupliStorageError(f'Failed to serialize episode: {str(e)}')

        try:
            with open(file_path, 'w', encoding='UTF-8') as f:
                f.write(serialized_object)

            # Check if the file was saved correctly
            if not file_path.exists():
                raise TupliStorageError(f'Failed to save episode to {file_path}')
            else:
                logger.info('Saved episode to %s', file_path)
        except Exception as e:
            raise TupliStorageError(f'Failed to write episode to file: {str(e)}')

        # Return the episode header
        return EpisodeHeader(
            **{k: v for k, v in episode_item.model_dump().items() if k != 'tuples'}
        )

    def publish_episode(self, uri: str, publish_in: str = 'global') -> None:
        """
        Publishes an episode in the specified group in local file storage.
        This is a placeholder to implement the interface.
        Args:
            uri (str): The ID of the episode to publish.
            publish_in (str): The group to publish the episode in. Defaults to 'global'.
        """
        logger.info('Publishing functionality is not available in FileStorage')

    def list_episodes(
        self, filter: BaseFilter = None, include_tuples: bool = False
    ) -> list[EpisodeHeader] | list[EpisodeItem]:
        """
        Lists all episodes in the local file storage that match the specified filter.

        Args:
            filter (BaseFilter, optional): The filter to apply when listing episodes.
            include_tuples (bool, optional): Whether to include tuples in the episode data.

        Returns:
            list[EpisodeHeader] | list[EpisodeItem]: A list of episode headers or full episode items.
        """
        episodes_dir = self.storage_dir / 'episodes'
        if not episodes_dir.exists():
            return []

        results = []
        filter_func = self.convert_filter_to_function(filter)

        for file_path in episodes_dir.glob('*.json'):
            try:
                with open(file_path, 'r', encoding='UTF-8') as f:
                    episode_dict = json.loads(f.read())

                # Apply filter function
                if filter_func(episode_dict):
                    # If we don't need to include tuples, create an EpisodeHeader
                    if not include_tuples:
                        # Create episode header object without tuples
                        header_dict = {k: v for k, v in episode_dict.items() if k != 'tuples'}
                        results.append(EpisodeHeader(**header_dict))
                    else:
                        # Include full episode with tuples
                        results.append(EpisodeItem(**episode_dict))

            except Exception as e:
                logger.info('Error loading episode from %s: %s', file_path, str(e))
                continue

        return results

    def delete_episode(self, uri: str) -> None:
        """
        Deletes an episode from the local file storage.

        Args:
            uri (str): The ID of the episode to delete.

        Raises:
            TupliStorageError: If the episode cannot be deleted.
        """
        # Construct file path from episode ID
        file_path = self.storage_dir / 'episodes' / f'{uri}.json'

        if not file_path.exists():
            raise TupliStorageError(f'Episode with ID {uri} does not exist.')

        try:
            file_path.unlink()
            logger.info('Deleted episode with ID %s', uri)
        except Exception as e:
            raise TupliStorageError(f'Failed to delete episode {uri}: {str(e)}')



import asyncio
import os

from pytupli.schema import (
    RIGHT_BUNDLE_CONTENT_CREATE,
    RIGHT_BUNDLE_CONTENT_DELETE,
    RIGHT_BUNDLE_CONTENT_READ,
    DEFAULT_ROLE,
    RIGHT,
    Membership,
    User,
)
from pytupli.server.config import USER_COLLECTION_NAME, USER_ROLES_COLLECTION_NAME, DBHandlerFactory
from pytupli.server.db.db_handler import MongoDBHandler
from pytupli.server.management.authorization import hash_password


default_roles = [
    {
        'role': DEFAULT_ROLE.ADMIN,
        'rights': [
            *RIGHT_BUNDLE_CONTENT_READ,
            *RIGHT_BUNDLE_CONTENT_CREATE,
            *RIGHT_BUNDLE_CONTENT_DELETE,
            RIGHT.USER_CREATE,
            RIGHT.USER_READ,
            RIGHT.USER_UPDATE,
            RIGHT.USER_DELETE,
            RIGHT.GROUP_CREATE,
            RIGHT.GROUP_READ,
            RIGHT.GROUP_UPDATE,
            RIGHT.GROUP_DELETE,
            RIGHT.ROLE_MANAGEMENT,
        ],
        'description': 'Full access to all resources and management functions.',
    },
    {
        'role': DEFAULT_ROLE.CONTENT_ADMIN,
        'rights': [
            *RIGHT_BUNDLE_CONTENT_READ,
            *RIGHT_BUNDLE_CONTENT_CREATE,
            *RIGHT_BUNDLE_CONTENT_DELETE,
        ],
        'description': 'Can create, edit, and delete content.',
    },
    {
        'role': DEFAULT_ROLE.USER_ADMIN,
        'rights': [
            RIGHT.USER_CREATE,
            RIGHT.USER_READ,
            RIGHT.USER_UPDATE,
            RIGHT.USER_DELETE,
        ],
        'description': 'Can manage users.',
    },
    {
        'role': DEFAULT_ROLE.GROUP_ADMIN,
        'rights': [
            RIGHT.GROUP_CREATE,
            RIGHT.GROUP_READ,
            RIGHT.GROUP_UPDATE,
            RIGHT.GROUP_DELETE,
            RIGHT.USER_READ,
        ],
        'description': 'Can manage groups.',
    },
    {
        'role': DEFAULT_ROLE.GUEST,
        'rights': RIGHT_BUNDLE_CONTENT_READ,
        'description': 'Limited access for guest users.',
    },
    {
        'role': DEFAULT_ROLE.CONTRIBUTOR,
        'rights': [
            *RIGHT_BUNDLE_CONTENT_CREATE,
            *RIGHT_BUNDLE_CONTENT_READ,
        ],
        'description': 'Can create and read content.',
    },
    {
        'role': DEFAULT_ROLE.MEMBER,
        'rights': [
            RIGHT.GROUP_READ,
            RIGHT.USER_READ,
            *RIGHT_BUNDLE_CONTENT_READ,
        ],
        'description': 'Can create and read content in a group.',
    },
    {
        'role': DEFAULT_ROLE.GLOBAL_MEMBER,
        'rights': [
            RIGHT.GROUP_CREATE,
            RIGHT.USER_READ,
            *RIGHT_BUNDLE_CONTENT_READ,
        ],
        'description': 'Can read content and create groups.',
    },
]


# Async function for creating roles in the database
async def initialize_roles(db_handler: MongoDBHandler):
    """
    Function initializes the roles in the database.
    """
    r = await db_handler.get_items(USER_ROLES_COLLECTION_NAME, {})
    existing_roles = [role['role'] for role in r]
    all_roles_exist = False
    for role in default_roles:
        if role['role'] not in existing_roles:
            all_roles_exist = False
            break
        all_roles_exist = True
    if all_roles_exist:
        print('Roles already exist')
        return
    res = await db_handler.create_items(USER_ROLES_COLLECTION_NAME, default_roles)
    print(res)


async def create_admin(db_handler: MongoDBHandler, admin_pw: str):
    """
    Function creates an admin user in the database.
    """
    r = await db_handler.get_item(USER_COLLECTION_NAME, {'username': 'admin'})
    if r:
        print('Admin user already exists')
        return
    hashed_password = hash_password(admin_pw)
    res = await db_handler.create_item(
        collection=USER_COLLECTION_NAME,
        item=User(
            username='admin',
            password=hashed_password,
            memberships=[Membership(group='global', roles=[DEFAULT_ROLE.ADMIN.value])],
        ).model_dump(),
    )
    print(res)


async def initialize_database(db_handler: DBHandlerFactory):
    """
    Function initializes the database with roles and an admin user.
    """
    if os.getenv('DB_ADMIN_PW') is None:
        raise ValueError(
            'DB_ADMIN_PW not set in environment. Please set this environment variable to continue.'
        )
    admin_pw = os.getenv('DB_ADMIN_PW')
    await initialize_roles(db_handler)
    await create_admin(db_handler, admin_pw)


# Run the async function
if __name__ == '__main__':
    from dotenv import load_dotenv

    load_dotenv(override=True)

    db_handler = DBHandlerFactory.get_handler()
    asyncio.run(initialize_database(db_handler))
    print('Verified roles and admin user')
